""" Control module for Peltier cooler via a TEC05-24 or TEC16-24 controller """
from __future__ import annotations

import aioserial
import pint
import numpy as np
import asyncio
from typing import List, Tuple
from dataclasses import dataclass

from loguru import logger
from asyncio import Lock
from flowchem import ureg
from flowchem.components.device_info import DeviceInfo
from flowchem.components.technical.temperature import TempRange
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.custom.peltier_cooler_component import (
    PeltierCoolerTemperatureControl,
)
from flowchem.utils.exceptions import InvalidConfigurationError
from flowchem.utils.people import jakob, miguel


class PeltierException(Exception):
    """ General peltier exception """
    pass


class InvalidConfiguration(PeltierException):
    """ Used for failure in the serial communication """
    pass


class InvalidCommand(PeltierException):
    """ The provided command is invalid. This can be caused by the peltier state e.g. if boundary values are prohibitive! """
    pass


class InvalidArgument(PeltierException):
    """ A valid command was followed by an invalid argument, usually out of accepted range """

    pass


class UnachievableSetpoint(PeltierException):
    """ A valid command was followed by an invalid argument, Out of hardware capabilities """

    pass


@dataclass
class PeltierCommandTemplate:
    """ Class representing a peltier command and its expected reply, but without target peltier number """

    command_string: str
    reply_lines: int  # Reply line without considering leading newline and tailing prompt!
    requires_argument: bool

    def to_peltier(self, address: int, argument: int | str = "") -> PeltierCommand:
        """ Returns a Command by adding to the template peltier address and command arguments """
        if self.requires_argument and not argument:
            raise InvalidArgument(
                f"Cannot send command {self.command_string} without an argument!"
            )
        elif self.requires_argument is False and argument:
            raise InvalidArgument(
                f"Cannot provide an argument to command {self.command_string}!"
            )
        return PeltierCommand(
            command_string=self.command_string,
            reply_lines=self.reply_lines,
            requires_argument=self.requires_argument,
            target_peltier_address=address,
            command_argument=str(argument),
        )


@dataclass
class PeltierCommand(PeltierCommandTemplate):
    """ Class representing a peltier command and its expected reply """

    target_peltier_address: int
    command_argument: str

    def compile(self, ) -> str:
        """
        Create actual command byte by prepending peltier address to command.
        """
        assert 0 <= self.target_peltier_address < 99
        # end character needs to be '\n'.
        if self.command_argument:
            return (
                str(self.target_peltier_address)
                + " "
                + self.command_string
                + " "
                + self.command_argument
                + "\n"
            )
        else:
            return (
                str(self.target_peltier_address)
                + " "
                + self.command_string
                + "\n"
            )


class PeltierIO:
    """ Setup with serial parameters, low level IO"""

    DEFAULT_CONFIG = {
        "timeout": 0.1,
        "baudrate": 115200,
        "parity": aioserial.PARITY_NONE,
        "stopbits": aioserial.STOPBITS_ONE,
        "bytesize": aioserial.EIGHTBITS,
    }

    # noinspection PyPep8
    def __init__(self, aio_port: aioserial.Serial):
        """Initialize communication on the serial port where the cooler is connected.

        Args:
        ----
            aio_port: aioserial.Serial() object
        """
        self.lock = Lock()
        self._serial = aio_port

    @classmethod
    def from_config(cls, port, **serial_kwargs):
        """Create PeltierIO from config."""
        # Merge default serial settings with provided ones.
        configuration = dict(PeltierIO.DEFAULT_CONFIG, **serial_kwargs)

        try:
            serial_object = aioserial.AioSerial(port, **configuration)
        except aioserial.SerialException as serial_exception:
            raise InvalidConfigurationError(
                f"Could not open serial port {port} with configuration {configuration}"
            ) from serial_exception

        return cls(serial_object)

    async def _write(self, command: PeltierCommand):
        """ Writes a command to the peltier """
        command_compiled = command.compile()
        logger.debug(f"Sending {repr(command_compiled)}")
        try:
            await self._serial.write_async(command_compiled.encode("ascii"))
        except aioserial.SerialException as e:
            raise InvalidConfiguration from e

    async def _read_reply(self, command) -> str:
        """ Reads the peltier reply from serial communication """
        logger.debug(
            f"I am going to read {command.reply_lines} line for this command (+prompt)"
        )
        reply_string = ""

        for line_num in range(
            command.reply_lines + 2
        ):  # +1 for leading newline character in reply + 1 for prompt
            chunk = await self._serial.readline_async()
            chunk = chunk.decode("ascii")
            logger.debug(f"Read line: {repr(chunk)} ")

            # Stripping newlines etc allows to skip empty lines and clean output
            chunk = chunk.strip()

            if chunk:
                reply_string += chunk

        logger.debug(f"Reply received: {reply_string}")
        return reply_string

    @staticmethod
    def parse_response_line(line: str) -> Tuple[int, str, str]:
        """ Split a received line in its components: address, prompt and reply body """
        assert len(line) > 0

        peltier_address = int(line.split(" ")[0])
        status = str(line.split(" ")[1].split("=")[0])
        reply = str(line.split(" ")[1].split("=")[1])
        return peltier_address, status, str(reply)

    @staticmethod
    def check_for_errors(last_response_line, command_sent):
        """ Further response parsing, checks for error messages """
        if "COMMAND ERR" in last_response_line:
            raise InvalidCommand(
                f"The command {command_sent} is invalid for Peltier {command_sent.target_peltier_address}!"
                f"[Reply: {last_response_line}]"
            )
        elif "NUMBER ERR" in last_response_line:
            raise InvalidArgument(
                f"The argument {command_sent} is out of allowed range for {command_sent.target_peltier_address}!"
                f"[Reply: {last_response_line}]"
            )
        elif "FORMAT ERR" in last_response_line:
            raise UnachievableSetpoint(
                f"The command {command_sent} to peltier {command_sent.target_peltier_address} is of invalid format, this likely means out of global boundaries"
                f"[Reply: {last_response_line}]"
            )


    def reset_buffer(self):
        """ Reset input buffer before reading from serial. In theory not necessary if all replies are consumed... """
        try:
            self._serial.reset_input_buffer()
        except aioserial.SerialException as e:
            raise InvalidConfiguration from e

    async def write_and_read_reply(
        self, command: PeltierCommand
    ) -> str:
        """ Main PeltierIO method. Sends a command to the peltier, read the replies and returns it, optionally parsed """
        async with self.lock:
            self.reset_buffer()
            await self._write(command)
            response = await self._read_reply(command)

        if not response:
            raise InvalidConfiguration(
                f"No response received from peltier, check peltier address! "
                f"(Currently set to {command.target_peltier_address})"
            )

        PeltierIO.check_for_errors(last_response_line=response, command_sent=command)

        # Parse reply
        peltier_address, return_status, parsed_response = PeltierIO.parse_response_line(response)

        # Ensures that all the replies came from the target peltier (this should always be the case)
        assert all(address == command.target_peltier_address for address in [peltier_address])

        return parsed_response


# noinspection SpellCheckingInspection
class PeltierCommands:

    """Holds the commands and arguments. """
    EMPTY_MESSAGE = PeltierCommandTemplate(
        command_string="", reply_lines=1, requires_argument=False
    )
    # TEMP1=-8.93 C
    GET_TEMPERATURE = PeltierCommandTemplate(
        command_string="GT1", reply_lines=1, requires_argument=False
    )
    # TEMP2 = -7.77C
    GET_SINK_TEMPERATURE = PeltierCommandTemplate(
        command_string="GT2", reply_lines=1, requires_argument=False
    )
    # TEMP_SET = 10.00 C ONLY WORKS IF ON
    SET_TEMPERATURE = PeltierCommandTemplate(
        command_string="STV", reply_lines=1, requires_argument=True
    )
    SET_SLOPE = PeltierCommandTemplate(
        command_string="STS", reply_lines=1, requires_argument=True
    )
    # STATUS=1
    SWITCH_ON = PeltierCommandTemplate(
        command_string="SEN", reply_lines=1, requires_argument=False
    )
    # STATUS=0
    SWITCH_OFF = PeltierCommandTemplate(
        command_string="SDI", reply_lines=1, requires_argument=False
    )
    COOLING_CURRENT_LIMIT = PeltierCommandTemplate(
        command_string="SCC", reply_lines=1, requires_argument=True
    )
    HEATING_CURRENT_LIMIT = PeltierCommandTemplate(
        command_string="SHC", reply_lines=1, requires_argument=True
    )
    SET_DIFFERENTIAL_PID = PeltierCommandTemplate(
        command_string="SDF", reply_lines=1, requires_argument=True
    )
    SET_INTEGRAL_PID = PeltierCommandTemplate(
        command_string="SIF", reply_lines=1, requires_argument=True
    )
    SET_PROPORTIONAL_PID = PeltierCommandTemplate(
        command_string="SPF", reply_lines=1, requires_argument=True
    )
    SET_T_MAX = PeltierCommandTemplate(
        command_string="SMA", reply_lines=1, requires_argument=True
    )
    SET_T_MIN = PeltierCommandTemplate(
        command_string="SMI", reply_lines=1, requires_argument=True
    )
    GET_POWER = PeltierCommandTemplate(
        command_string="GCU", reply_lines=1, requires_argument=False
    )
    GET_CURRENT = PeltierCommandTemplate(
        command_string="GPW", reply_lines=1, requires_argument=False
    )
    GET_SETTINGS = PeltierCommandTemplate(
        command_string="GPA", reply_lines=1, requires_argument=False
    )


class PeltierDefaults:
    HEATING_PID = [0.64, 0.53, 0.13]
    COOLING_PID = [2.83, 2.36, 0.59]
    BASE_TEMP = -7.6
    state_dependent_data: List[List[float]] = [[-55, 50], [14, 14], [10, 10]]
    STATE_DEPENDANT_CURRENT_LIMITS = np.array(state_dependent_data, dtype=float).transpose()
    T_MAX = 50
    T_MIN = -55


class PeltierLowCoolingDefaults(PeltierDefaults):
    HEATING_PID = [2, 0.03, 0]
    COOLING_PID = HEATING_PID
    BASE_TEMP = -24.7
    state_dependent_data: List[List[float]] = [
        [-65, -60, -55, -50, -40, -30, -20, -10, 0, 10, 20, 30],
        [7.5, 6.5, 5, 4, 3, 3, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0.5, 1, 1, 2.5, 3, 3.5, 3.5, 3.5]
    ]
    STATE_DEPENDANT_CURRENT_LIMITS = np.array(state_dependent_data, dtype=float).transpose()
    T_MAX = 30
    T_MIN = -66


class PeltierCooler(FlowchemDevice):
    """Peltier Cooler module class."""

    def __init__(
            self,
            peltier_io: PeltierIO,
            name: str = "",
            address: int = 0,
            peltier_defaults: str | None = None,
    ) -> None:
        super().__init__(name)
        self.peltier_io = peltier_io
        self.address: int = address
        match peltier_defaults:
            case None | "default":
                self.peltier_defaults = PeltierDefaults()
            case "low_cooling":
                self.peltier_defaults = PeltierLowCoolingDefaults()

        # ToDo check info
        self.device_info = DeviceInfo(
            authors=[jakob, miguel],
            manufacturer="Custom",
            model="Custom",
        )

    @classmethod
    def from_config(
            cls,
            port: str,
            address: int,
            name: str = "",
            peltier_defaults: str | None = None,
            **serial_kwargs,
    ):

        peltier_io = PeltierIO.from_config(port, **serial_kwargs)

        return cls(peltier_io=peltier_io, address=address, name=name, peltier_defaults=peltier_defaults)

    async def initialize(self):
        await self.set_default_values()
        await self.set_pid_parameters(*self.peltier_defaults.COOLING_PID)

        temp_range = TempRange(
            min=ureg.Quantity(f"{self.peltier_defaults.T_MIN} °C"),
            max=ureg.Quantity(f"{self.peltier_defaults.T_MAX} °C")
        )
        list_of_components = [
            PeltierCoolerTemperatureControl("temperature_control", self, temp_limits=temp_range)
        ]
        self.components.extend(list_of_components)
        logger.info(
            f"Connected to peltier on port {self.peltier_io._serial.port}:{self.address}!")

    async def set_default_values(self):
        await self._set_max_temperature(self.peltier_defaults.T_MAX)
        await self._set_min_temperature(self.peltier_defaults.T_MIN)
        await self._set_current_limit_heating(int(self.peltier_defaults.STATE_DEPENDANT_CURRENT_LIMITS[-1::, 2]))
        await self._set_current_limit_cooling(int(self.peltier_defaults.STATE_DEPENDANT_CURRENT_LIMITS[:1:, 1]))
        await self.disable_slope()

    async def set_pid_parameters(self, proportional, integral, differential):
        await self._set_p_of_pid(proportional)
        await self._set_i_of_pid(integral)
        await self._set_d_of_pid(differential)

    async def send_command_and_read_reply(self,
                                          command_template: PeltierCommandTemplate,
                                          parameter: int | str = "",
                                          ) -> str:
        """ Sends a command based on its template and return the corresponding reply as str """
        return await self.peltier_io.write_and_read_reply(command_template.to_peltier(self.address, str(parameter)))

    async def set_temperature(self, temperature: pint.Quantity):
        await self.stop_control()
        # start softly
        await self._set_current_limit_cooling(0.5)
        await self._set_current_limit_heating(0.5)
        await self._set_temperature(temperature.m_as("°C"))
        await self.start_control()
        await asyncio.sleep(10)
        # Now start with power
        await self.set_default_values()
        await self._set_state_dependant_parameters(temperature.m_as("°C"))

    async def _set_temperature(self, temperature: float):
        reply = await self.send_command_and_read_reply(PeltierCommands.SET_TEMPERATURE, round(temperature * 100))
        assert float(reply) == temperature

    async def set_slope(self, slope: float):
        reply = await self.send_command_and_read_reply(PeltierCommands.SET_SLOPE, round(slope * 100))
        assert float(reply) == slope

    async def disable_slope(self):
        reply = await self.send_command_and_read_reply(PeltierCommands.SET_SLOPE, 0)
        assert int(reply) == 0

    async def start_control(self):
        reply = await self.send_command_and_read_reply(PeltierCommands.SWITCH_ON)
        assert int(reply) == 1

    async def get_temperature(self) -> float:
        reply = await self.send_command_and_read_reply(PeltierCommands.GET_TEMPERATURE)
        return float(reply)

    async def get_sink_temperature(self) -> float:
        reply = await self.send_command_and_read_reply(PeltierCommands.GET_SINK_TEMPERATURE)
        return float(reply)

    async def stop_control(self):
        reply = await self.send_command_and_read_reply(PeltierCommands.SWITCH_OFF)
        assert int(reply) == 0

    async def go_to_rt_and_switch_off(self):
        # set to RT, wait 2 min, stop T-control: This is just a convenience and safety measure: if the Peltier is
        # shut off and the heating is directly shut off, the heating might freeze
        await self.set_temperature(ureg.Quantity("25 °C"))
        await asyncio.sleep(120)
        await self.stop_control()

    async def get_power(self) -> float:
        # return power in W
        reply = await self.send_command_and_read_reply(PeltierCommands.GET_POWER)
        return float(reply)

    async def get_current(self) -> int:
        # return power in W
        reply = await self.send_command_and_read_reply(PeltierCommands.GET_CURRENT)
        return int(reply)

    async def get_parameters(self) -> str:
        # return parameter list
        reply = await self.send_command_and_read_reply(PeltierCommands.GET_SETTINGS)
        return reply

    async def _set_current_limit_cooling(self, current_limit: float):
        # current in amp
        reply = await self.send_command_and_read_reply(PeltierCommands.COOLING_CURRENT_LIMIT, round(current_limit * 100))
        assert float(reply) == current_limit

    async def _set_current_limit_heating(self, current_limit: float):
        # current in amp
        reply = await self.send_command_and_read_reply(PeltierCommands.HEATING_CURRENT_LIMIT, round(current_limit * 100))
        assert float(reply) == current_limit

    async def _set_d_of_pid(self, differential: float):
        # max 10
        reply = await self.send_command_and_read_reply(PeltierCommands.SET_DIFFERENTIAL_PID, round(differential * 100))
        assert float(reply) == differential

    async def _set_i_of_pid(self, integral):
        # max 10
        reply = await self.send_command_and_read_reply(PeltierCommands.SET_INTEGRAL_PID, round(integral * 100))
        assert float(reply) == integral

    async def _set_p_of_pid(self, proportional):
        # max 10
        reply = await self.send_command_and_read_reply(PeltierCommands.SET_PROPORTIONAL_PID, round(proportional * 100))
        assert float(reply) == proportional

    async def _set_max_temperature(self, t_max):
        # max 10
        reply = await self.send_command_and_read_reply(PeltierCommands.SET_T_MAX, round(t_max * 100))
        assert float(reply) == t_max

    async def _set_min_temperature(self, t_min):
        # max 10
        reply = await self.send_command_and_read_reply(PeltierCommands.SET_T_MIN, round(t_min * 100))
        assert float(reply) == t_min

    async def _set_state_dependant_parameters(self, new_T_setpoint):
        if self.peltier_defaults.BASE_TEMP < new_T_setpoint:
            # set_heating_parameters
            await self.set_pid_parameters(*self.peltier_defaults.HEATING_PID)
        else:
            await self.set_pid_parameters(*self.peltier_defaults.COOLING_PID)
        if new_T_setpoint > self.peltier_defaults.BASE_TEMP:
            # set current limit for heating
            n = np.where((self.peltier_defaults.STATE_DEPENDANT_CURRENT_LIMITS[::, 0] >= new_T_setpoint))[0][0]
            settings = self.peltier_defaults.STATE_DEPENDANT_CURRENT_LIMITS[n, :]

        else:
            # set to cooling
            n = np.where((self.peltier_defaults.STATE_DEPENDANT_CURRENT_LIMITS[::, 0] <= new_T_setpoint))[0][-1]
            settings = self.peltier_defaults.STATE_DEPENDANT_CURRENT_LIMITS[n, :]
        await self._set_current_limit_cooling(float(settings[1]))
        await self._set_current_limit_heating(float(settings[2]))




from __future__ import annotations

from typing import TYPE_CHECKING

from flowchem.devices.flowchem_device import FlowchemDevice

if TYPE_CHECKING:
    from .pressure_sensor import PhidgetPressureSensor
from flowchem.components.sensors.pressure_sensor import PressureSensor


class PhidgetPressureSensorComponent(PressureSensor):
    hw_device: PhidgetPressureSensor  # just for typing

    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        """
        Initialize the PhidgetPressureSensorComponent with a name and hardware device.

        Args:
            name (str): The name of the sensor component.
            hw_device (FlowchemDevice): The hardware device associated with this sensor.
        """
        super().__init__(name, hw_device)

    async def read_pressure(self, units: str = "bar"):
        """
        Read the pressure from the sensor and return it in the specified units.

        Args:
            units (str): The units to express the pressure in. Default is "bar".

        Returns:
            float: The pressure reading expressed in the specified units.
        """
        return self.hw_device.read_pressure().m_as(units)



from __future__ import annotations

from typing import TYPE_CHECKING

from flowchem.components.technical.power import PowerControl

if TYPE_CHECKING:
    from flowchem.devices import MansonPowerSupply


class MansonPowerControl(PowerControl):
    hw_device: MansonPowerSupply  # for typing's sake

    async def set_current(self, current: str):
        """Set the target current using a "magnitude and unit" format string.

        Args:
            current (str): The desired current as a string in "magnitude and unit" format (e.g., '5 A', '500 mA').

        Returns:
            Awaitable: Result of the set_current operation from the hardware device.
        """
        return await self.hw_device.set_current(current)

    async def get_current(self) -> float:
        """Retrieve the current output in Amperes.

        Returns:
            float: The current output in Amperes.
        """
        return await self.hw_device.get_output_current()

    async def set_voltage(self, voltage: str):
        """Set the target voltage using a "magnitude and unit" format string.

        Args:
            voltage (str): The desired voltage as a string in "magnitude and unit" format (e.g., '12 V', '3.3 V').

        Returns:
            Awaitable: Result of the set_voltage operation from the hardware device.
        """
        return await self.hw_device.set_voltage(voltage)

    async def get_voltage(self) -> float:
        """Retrieve the current output voltage in Volts.

        Returns:
            float: The current output voltage in Volts.
        """
        return await self.hw_device.get_output_voltage()

    async def power_on(self):
        """Turn on the power supply output.

        Returns:
            Awaitable: Result of the power on operation from the hardware device.
        """
        return await self.hw_device.output_on()

    async def power_off(self):
        """Turn off the power supply output.

        Returns:
            Awaitable: Result of the power off operation from the hardware device.
        """
        return await self.hw_device.output_off()



"""Control module for the Vapourtec R2 valves."""
from __future__ import annotations

from typing import TYPE_CHECKING

from loguru import logger

from flowchem.components.pumps.hplc_pump import HPLCPump
from flowchem.components.sensors.sensor import Sensor
from flowchem.components.sensors.pressure_sensor import PressureSensor
from flowchem.components.technical.photo import Photoreactor
from flowchem.components.technical.power import PowerSwitch
from flowchem.components.technical.temperature import TemperatureControl, TempRange
from flowchem.components.valves.distribution_valves import TwoPortDistributionValve
from flowchem.components.valves.injection_valves import SixPortTwoPositionValve

if TYPE_CHECKING:
    from .r2 import R2

import json

AllValveDic = {
    "TwoPortValve_A": 0,
    "TwoPortValve_B": 1,
    "InjectionValve_A": 2,
    "InjectionValve_B": 3,
    "TwoPortValve_C": 4,
}
AllPumpDic = {"HPLCPump_A": 0, "HPLCPump_B": 1}


class R2GeneralSensor(Sensor):
    hw_device: R2  # for typing's sake

    def __init__(self, name: str, hw_device: R2) -> None:
        super().__init__(name, hw_device)
        self.add_api_route("/monitor-system", self.monitor_sys, methods=["GET"])
        self.add_api_route("/get-run-state", self.get_run_state, methods=["GET"])
        self.add_api_route(
            "/set-system-max-pressure",
            self.set_sys_pressure_limit,
            methods=["PUT"],
        )

    async def monitor_sys(self) -> dict:
        """Monitor system performance."""
        return await self.hw_device.pooling()

    async def get_run_state(self) -> str:
        """Get current system state."""
        return await self.hw_device.get_state()

    async def set_sys_pressure_limit(self, pressure: str) -> bool:
        """Set maximum system pressure: range 1,000 to 50,000 mbar."""
        # TODO: change to accept different units
        await self.hw_device.set_pressure_limit(pressure)
        return True


class R4Reactor(TemperatureControl):
    """R4 reactor heater channel controlled via the R2."""

    hw_device: R2  # for typing's sake

    def __init__(
        self, name: str, hw_device: R2, channel: int, temp_limits: TempRange
    ) -> None:
        """Create a TemperatureControl object."""
        super().__init__(name, hw_device, temp_limits)
        self.channel = channel

    async def set_temperature(self, temperature: str, heating: bool | None = None):
        """Set the target temperature to the given string in "magnitude and unit" format."""
        set_t = await super().set_temperature(temperature)
        return await self.hw_device.set_temperature(self.channel, set_t, heating)

    async def get_temperature(self) -> float:  # type: ignore
        """Return temperature in Celsius."""
        return await self.hw_device.get_current_temperature(self.channel)

    async def is_target_reached(self) -> bool | None:  # type: ignore
        """Return True if the set temperature target has been reached."""
        current_temp = await self.hw_device.get_current_temperature(self.channel)
        target_temp = await self.hw_device.get_target_temperature(self.channel)
        if target_temp == "-1000":
            return None
        elif abs(current_temp - target_temp) <= 2:
            return True
        else:
            return False

    async def power_on(self):
        """Turn on temperature control."""
        return await self.hw_device.power_on()

    async def power_off(self):
        """Turn off temperature control."""
        return await self.hw_device.power_off()


class UV150PhotoReactor(Photoreactor):
    """R2 reactor control class."""

    hw_device: R2  # for typing's sake

    def __init__(self, name: str, hw_device: R2) -> None:
        super().__init__(name, hw_device)
        # self._intensity = 0  # 0 set upon device init

    async def set_intensity(self, percent: int = 100):
        """Set UV light intensity at the range 50-100%."""
        self.hw_device._intensity = percent
        await self.hw_device.set_UV150(percent)

    async def get_intensity(self) -> int:
        """Return last set intensity."""
        return self.hw_device._intensity

    async def power_on(self):
        """Turn on the whole system, no way to power UV150 independently."""
        if self.hw_device._intensity:
            await self.hw_device.power_on()
            return
        logger.error("UV150 power on requested without setting intensity first!")

    async def power_off(self):
        """Turn off light."""
        self.hw_device._intensity = 0
        await self.hw_device.set_UV150(0)


class R2InjectionValve(SixPortTwoPositionValve):
    """R2 reactor injection loop valve control class."""

    hw_device: R2  # for typing's sake

    # get position
    position_mapping = {"load": "0", "inject": "1"}
    _reverse_position_mapping = {v: k for k, v in position_mapping.items()}

    def __init__(self, name: str, hw_device: R2, valve_code: int) -> None:
        """Create a ValveControl object."""
        super().__init__(name, hw_device)
        self.valve_code = valve_code

        self.add_api_route("/monitor_position", self.get_monitor_position, methods=["GET"])
        self.add_api_route("/monitor_position", self.set_monitor_position, methods=["PUT"])

    def _change_connections(self, raw_position, reverse: bool = False) -> str:
        return raw_position

    async def get_position(self) -> list[list]:
        """Get current valve position."""
        position = await self.hw_device.get_valve_position(self.valve_code)
        return self._positions[int(self._change_connections(position, reverse=True))]

    async def set_position(self,
                           connect: str = "",
                           disconnect: str = "",
                           ambiguous_switching: str | bool = False):
        """Move valve to position, which connects named ports. For example, [[5,0]] or [[2,3]]"""
        positions_to_connect_l = json.loads(connect)
        position_to_connect = tuple(tuple(inner_list) for inner_list in positions_to_connect_l)
        target_pos = str(self._connect_positions(position_to_connect))
        target_pos = self._change_connections(target_pos)
        await self.hw_device.trigger_key_press(
            str(self.valve_code * 2 + int(target_pos)),
        )
        return True

    async def get_monitor_position(self) -> str:
        """Get current valve position."""
        position = await self.hw_device.get_valve_position(self.valve_code)
        return "position is %s" % self._reverse_position_mapping[position]

    async def set_monitor_position(self, position: str) -> bool:
        """Set position: 'load' or 'inject'."""
        target_pos = self.position_mapping[position]  # load or inject
        await self.hw_device.trigger_key_press(
            str(self.valve_code * 2 + int(target_pos)),
        )
        return True


class R2TwoPortValve(TwoPortDistributionValve):  # total 3 positions (A, B, Collection)
    """R2 reactor injection loop valve control."""

    hw_device: R2

    # TODO: the mapping name is not applicable
    # TODO this needs to be adjusted to new code
    position_mapping = {"Solvent": "0", "Reagent": "1"}
    _reverse_position_mapping = {v: k for k, v in position_mapping.items()}

    def __init__(self, name: str, hw_device: R2, valve_code: int) -> None:
        """Create a ValveControl object."""
        super().__init__(name, hw_device)
        self.valve_code = valve_code
        # raise NotImplementedError("Check that the mapping is correct")
        self.add_api_route("/monitor_position", self.get_monitor_position, methods=["GET"])
        self.add_api_route("/monitor_position", self.set_monitor_position, methods=["PUT"])

    def _change_connections(self, raw_position, reverse: bool = False) -> str:
        if not reverse:
            translated = raw_position
        else:
            translated = raw_position
        return translated

    async def get_position(self) -> list[list]:
        """Get current valve position."""
        position = await self.hw_device.get_valve_position(self.valve_code)
        return (self._positions[int(self._change_connections(position, reverse=True))])

    async def set_position(self,
                           connect: str = "",
                           disconnect: str = "",
                           ambiguous_switching: str | bool = False):
        """Move valve to position, which connects named ports. For example, [[5,0]] or [[2,3]]"""
        positions_to_connect_l = json.loads(connect)
        position_to_connect = tuple(tuple(inner_list) for inner_list in positions_to_connect_l)
        target_pos = str(self._connect_positions(position_to_connect))
        target_pos = self._change_connections(target_pos)
        await self.hw_device.trigger_key_press(
            str(self.valve_code * 2 + int(target_pos)),
        )
        return True

    async def get_monitor_position(self) -> str:
        """Get current valve position."""
        position = await self.hw_device.get_valve_position(self.valve_code)
        # self.hw_device.last_state.valve[self.valve_number]
        return "inlet is %s" % self._reverse_position_mapping[position]

    async def set_monitor_position(self, position: str) -> bool:
        """Move valve to position."""
        target_pos = self.position_mapping[position]
        await self.hw_device.trigger_key_press(
            str(self.valve_code * 2 + int(target_pos)),
        )
        return True


class R2HPLCPump(HPLCPump):
    """R2 reactor injection loop valve control class."""

    hw_device: R2  # for typing's sake

    def __init__(self, name: str, hw_device: R2, pump_code: str) -> None:
        """Create a pump object."""
        super().__init__(name, hw_device)
        self.pump_code = pump_code

    async def get_current_flow(self) -> float:
        """Get current flow rate."""
        return await self.hw_device.get_current_flow(self.pump_code)

    async def set_flowrate(self, flowrate: str) -> bool:
        """Set flow rate to the pump."""
        await self.hw_device.set_flowrate(self.pump_code, flowrate)
        return True

    async def infuse(self, rate: str = "", volume: str = "") -> bool:
        """Set the flow rate: in ul/min and start infusion."""
        if volume:
            logger.warning(f"Volume parameter ignored: not supported by {self.name}!")

        await self.hw_device.set_flowrate(self.pump_code, rate)
        await self.hw_device.power_on()
        return True

    async def stop(self):
        """Stop infusion."""
        await self.hw_device.set_flowrate(pump=self.pump_code, flowrate="0 ul/min")

    async def is_pumping(self) -> bool:
        c_flow = await self.hw_device.get_current_flow(self.pump_code)
        return c_flow != 0


class R2PumpPressureSensor(PressureSensor):
    hw_device: R2  # for typing's sake

    def __init__(self, name: str, hw_device: R2, pump_code: int) -> None:
        """Create a ValveControl object."""
        super().__init__(name, hw_device)
        self.pump_code = pump_code

    async def read_pressure(self, units: str = "mbar") -> int | float | None:  # mbar
        """Get current pump pressure in mbar."""
        pressure = await self.hw_device.get_current_pressure(self.pump_code)
        return pressure.m_as(units)


class R2GeneralPressureSensor(PressureSensor):
    hw_device: R2  # for typing's sake

    def __init__(self, name: str, hw_device: R2) -> None:
        """Create a ValveControl object."""
        super().__init__(name, hw_device)

    async def read_pressure(self, units: str = "mbar") -> int:
        """Get system pressure."""
        pressure = await self.hw_device.get_current_pressure()
        return pressure.m_as(units)


class R2MainSwitch(PowerSwitch):
    hw_device: R2  # just for typing

    async def power_on(self) -> bool:
        await self.hw_device.power_on()
        return True

    async def power_off(self) -> bool:
        await self.hw_device.power_off()
        return True



"""CVC3000 Pressure Contol component."""
from __future__ import annotations

from typing import TYPE_CHECKING

from flowchem.components.technical.pressure import PressureControl
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.vacuubrand.constants import ProcessStatus, PumpState

if TYPE_CHECKING:
    from flowchem.devices.vacuubrand.cvc3000 import CVC3000


class CVC3000PressureControl(PressureControl):
    """
    Control component for managing pressure in the Vacuubrand CVC3000 vacuum system.

    This class provides methods to set and get pressure, check if the target pressure has been reached,
    and control the power state of the pressure control system.

    Attributes:
    -----------
    hw_device : CVC3000
        The hardware device instance of the CVC3000 vacuum controller.

    Methods:
    --------
    set_pressure(pressure: str) -> bool:
        Set the target pressure using a string representation - unit have to be provided.
    get_pressure() -> float:
        Retrieve the current pressure from the device in mbar.
    is_target_reached() -> bool:
        Check if the target pressure has been reached.
    power_on() -> bool:
        Turn on the pressure control.
    power_off() -> bool:
        Turn off the pressure control.
    """

    hw_device: CVC3000  # for typing's sake

    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        """
        Initialize the CVC3000PressureControl component.

        Parameters:
        -----------
        name : str
            The name assigned to this pressure control component.
        hw_device : FlowchemDevice
            The hardware device instance associated with this pressure control component.
        """
        super().__init__(name, hw_device)

        self.add_api_route(
            "/status",
            self.hw_device.status,
            response_model=ProcessStatus,
            methods=["PUT"],
        )

    async def set_pressure(self, pressure: str):
        """
        Set the target pressure on the CVC3000 device.

        Parameters:
        -----------
        pressure : str
            The target pressure to be set, expressed in a string format (e.g., mbar).
            For example: "500 mbar"

        Returns:
        --------
        bool
            Returns True if the operation was successful.
        """
        set_p = await super().set_pressure(pressure)
        return await self.hw_device.set_pressure(set_p)

    async def get_pressure(self) -> float:
        """
        Retrieve the current pressure from the CVC3000 device.

        Returns:
        --------
        float
            The current pressure in mbar.
        """
        return await self.hw_device.get_pressure()

    async def is_target_reached(self) -> bool:
        """
        Check if the target pressure has been reached.

        Returns:
        --------
        bool
            Returns True if the target pressure has been reached, False otherwise.
        """
        status = await self.hw_device.status()
        return status.state == PumpState.VACUUM_REACHED

    async def power_on(self) -> str:
        """
        Turn on the pressure control.

        Returns:
        --------
        str
            Returns binary string if the command to start the pressure control was successful.
        """
        return await self.hw_device._send_command_and_read_reply("START")

    async def power_off(self):
        """
        Turn off the pressure control.

        Returns:
        --------
        str
            Returns binary string if the command to stop the pressure control was successful.
        """
        return await self.hw_device._send_command_and_read_reply("STOP")



"""Control module for the Knauer DAD."""
import asyncio
from typing import TYPE_CHECKING

from loguru import logger

from flowchem.devices.flowchem_device import RepeatedTaskInfo
from flowchem.components.device_info import DeviceInfo
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.knauer._common import KnauerEthernetDevice
from flowchem.devices.knauer.dad_component import (
    DADChannelControl,
    KnauerDADLampControl,
)
from flowchem.utils.exceptions import InvalidConfigurationError
from flowchem.utils.people import dario, jakob, wei_hsin

if TYPE_CHECKING:
    pass

try:
    from flowchem_knauer import KnauerDADCommands

    HAS_DAD_COMMANDS = True
except ImportError:
    HAS_DAD_COMMANDS = False


class KnauerDAD(KnauerEthernetDevice, FlowchemDevice):
    """DAD control class."""

    def __init__(
            self,
            ip_address: object = None,
            mac_address: object = None,
            name: str | None = None,
            turn_on_d2: bool = False,
            turn_on_halogen: bool = False,
            display_control: bool = True,
    ) -> None:
        super().__init__(ip_address, mac_address, name=name)
        self.eol = b"\n\r"
        self._d2 = turn_on_d2
        self._hal = turn_on_halogen
        self._state_d2 = False
        self._state_hal = False
        self._control = display_control  # True for Local

        if not HAS_DAD_COMMANDS:
            raise InvalidConfigurationError(
                "You tried to use a Knauer DAD device but the relevant commands are missing!\n"
                "Unfortunately, we cannot publish those as they were provided under NDA.\n"
                "Contact Knauer for further assistance."
            )

        self.cmd = KnauerDADCommands()
        self.device_info = DeviceInfo(
            authors=[dario, jakob, wei_hsin],
            manufacturer="Knauer",
            model="DAD",
        )

    async def initialize(self):
        """Initialize connection."""
        await super().initialize()

        if self._control:
            await self.display_control(True)

        # get device information
        logger.info(f"Connected with Knauer DAD num {await self.serial_num()}")
        logger.info(f"Knauer DAD info: {await self.identify()} {await self.info()}")
        logger.info(f"Knauer DAD status: {await self.status()}")

        await self.set_wavelength(1, 254)
        await self.bandwidth(8)
        logger.info("set channel 1 : WL = 254 nm, BW = 8nm ")

        self.components = [
            KnauerDADLampControl("d2", self),
            KnauerDADLampControl("hal", self),
        ]

        self.components.extend(
            [DADChannelControl(f"channel{n + 1}", self, n + 1) for n in range(4)]
        )

    async def lamp(self, lamp: str, state: bool | str = "REQUEST") -> str:
        """Turn on or off the lamp, or request lamp state."""
        if isinstance(state, bool):
            state = "ON" if state else "OFF"

        lamp_mapping = {"d2": "_D2", "hal": "_HAL"}

        lampstatus_mapping = {
            "REQUEST": "?",
            "OFF": "0",
            "ON": "1",
            "HEAT": "2",
            "ERROR": "3",
        }
        _reverse_lampstatus_mapping = {v: k for k, v in lampstatus_mapping.items()}

        cmd = self.cmd.LAMP.format(
            lamp=lamp_mapping[lamp],
            state=lampstatus_mapping[state],
        )
        response = await self._send_and_receive(cmd)  # 'LAMP_D2:0'
        return response
        # if response.isnumeric() else _reverse_lampstatus_mapping[response[response.find(":") + 1:]]

    async def serial_num(self) -> str:
        """Get serial number."""
        return await self._send_and_receive(self.cmd.SERIAL)

    async def identify(self) -> str:
        """Get the instrument information
        CATEGORY (=3), MANUFACTURER,  MODEL_NR, SERNUM, VERSION,  MODIFICATION
        Example: 3,KNAUER,PDA-1,CSA094400001,2,01.
        """
        return await self._send_and_receive(self.cmd.IDENTIFY)

    async def info(self) -> str:
        """Get the instrument information
        NUMBER OF PIXEL (256, 512, 1024), SPECTRAL RANGE(“UV”, “VIS”, “UV-VIS”),
        HARDWARE VERSION, YEAR OF PRODUCTION,WEEK OF PRODUCTION,,CALIBR. A,CALIBR. B,, CALIBR. C.
        """
        return await self._send_and_receive(self.cmd.INFO)

    async def status(self):
        """Get status of the instrument
        Sending spectra (ON = 1, OFF = 0),
        D2 Lamp (OFF = 0, ON = 1, HEAT= 2, ERROR = 3),
        HAL Lamp (OFF = 0, ON = 1, ERROR = 3),
        Shutter(OFF = 0, ON=1, FILTER=2),
        External Error IN, External Start IN, External Autozero IN,
        Event1 OUT, Event2 OUT, Event3 OUT, Valve OUT, Error Code.
        """
        return await self._send_and_receive(self.cmd.STATUS)

    async def display_control(self, control: bool = True):
        cmd = self.cmd.LOCAL if control else self.cmd.REMOTE
        self._control = control
        return await self._send_and_receive(cmd)

    async def shutter(self, shutter: str) -> str:
        shutter_mapping = {"REQUEST": "?", "CLOSED": "0", "OPEN": "1", "FILTER": "2"}
        _reverse_shutter_mapping = {v: k for k, v in shutter_mapping.items()}

        cmd = self.cmd.SHUTTER.format(state=shutter_mapping[shutter])
        response = await self._send_and_receive(cmd)
        return (
            response
            if not response.isnumeric()
            else _reverse_shutter_mapping[response[response.find(":") + 1:]]
        )

    async def signal_type(self, s_type: str = "microAU") -> str:
        """Set and get the type of signal shown on the display
        0 = signal is Absorption Units
        1 = signal is intensity.
        """
        type_mapping = {"REQUEST": "?", "microAU": "0", "intensity": "1"}
        _reverse_type_mapping = {v: k for k, v in type_mapping.items()}

        cmd = self.cmd.SIGNAL_TYPE.format(state=type_mapping[s_type])
        response = await self._send_and_receive(cmd)
        return (
            response
            if not response.isnumeric()
            else _reverse_type_mapping[response[response.find(":") + 1:]]
        )

    async def get_wavelength(self, channel: int) -> int:
        cmd = self.cmd.WAVELENGTH.format(channel=channel, wavelength="?")
        return int(await self._send_and_receive(cmd))

    async def set_wavelength(self, channel: int, wavelength: int) -> str:
        """Set and read wavelength."""
        cmd = self.cmd.WAVELENGTH.format(channel=channel, wavelength=wavelength)
        return await self._send_and_receive(cmd)

    async def set_signal(self, channel: int, signal: int = 0):
        """Set signal to specific number."""
        cmd = self.cmd.SIGNAL.format(channel=channel, signal=signal)
        return await self._send_and_receive(cmd)

    async def read_signal(self, channel: int) -> float:
        """Read signal
        -9999999 to +9999999 (μAU, SIG_SRC = 0); 0 to 1000000 (INT, SIG_SRC = 1).
        """
        cmd = self.cmd.SIGNAL.format(channel=channel, signal="?")
        response = await self._send_and_receive(cmd)
        logger.info(f"signal: {response}")  # SIG1:113545,
        parsed_response = response.split(":")
        if parsed_response[1] == "OK":
            logger.warning(
                f"ValueError[channel{channel}]:the reply of 'get signal' command is OK.."
            )
            return await self.read_signal(channel)
        elif parsed_response[0] == f"SIG{channel}":
            return float(parsed_response[1]) / 10000
        else:
            logger.warning(
                f"ValueError[channel{channel}]: receive the reply not for channel{channel}!"
            )
            return await self.read_signal(channel)

    async def integration_time(self, integ_time: int | str = "?") -> str | int:
        """Set and read the integration time in 10 - 2000 ms."""
        cmd = self.cmd.INTEGRATION_TIME.format(time=integ_time)
        response = await self._send_and_receive(cmd)
        try:
            return int(response)  # ms
        except ValueError:
            return response

    async def bandwidth(self, bw: str | int) -> str | int:
        """Set bandwidth in the range of 4 to 25 nm
        read the setting of bandwidth.
        """
        if isinstance(bw, int):
            cmd = self.cmd.BANDWIDTH.format(bandwidth=bw)
            return await self._send_and_receive(cmd)
        else:
            cmd = self.cmd.BANDWIDTH.format(bandwidth="?")
            response = await self._send_and_receive(cmd)
            return int(response)

    def repeated_task(self):
        async def keepalive():
            await self.status()

        return RepeatedTaskInfo(seconds_every=45, task=keepalive)


async def main(dad):
    """Test function."""
    await dad.initialize()
    lamp_d2, lamp_hal, ch1, ch2, ch3, ch4 = dad.components()
    # bg1 = dad.bg_keep_connect()
    dad.info()
    # await asyncio.gather(asyncio.to_thread(bg2), bg2)

    # # set signal of channel 1 to zero
    # # await DAD.set_signal(1)
    # await ch1.set_wavelength(520)
    # await ch1.set_integration_time(70)
    # await ch1.set_bandwidth(4)
    # await ch1.acquire_signal()
    # await asyncio.sleep(60)
    # await DAD.initialize()
    # await ch1.acquire_signal()

    # set signal of channel 1 to zero
    # await DAD.set_signal(1)
    await ch1.set_wavelength(500)
    await ch1.set_integration_time(70)
    await ch1.set_bandwidth(8)
    await ch1.acquire_signal()


if __name__ == "__main__":
    k_dad = KnauerDAD(
        ip_address="192.168.10.7", turn_on_d2=False, turn_on_halogen=False
    )
    asyncio.run(main(k_dad))



"""ML600 component relative to valve switching."""
from __future__ import annotations

from typing import TYPE_CHECKING

from flowchem.components.valves.distribution_valves import ThreePortFourPositionValve, FourPortFivePositionValve

if TYPE_CHECKING:
    from .ml600 import ML600


class ML600LeftValve(FourPortFivePositionValve):
    """
    Represents the left valve of the ML600 pump with specific translation for raw positions.

    This valve has 8 possible positions each separated by 45 degrees, but only a few are functional.

    # 0 degree syr-left,
    # 45 right-front
    # 90 nothing
    # 135 front-syr
    # 180
    # 225 left front
    # 270 syr-right
    # 315
    # 360

    Attributes:
    -----------
    hw_device : ML600
        The hardware device instance associated with this valve.
    identifier : str
        The identifier for this valve, set to "B".

    Methods:
    --------
    _change_connections(raw_position: int, reverse: bool = False) -> int:
        Translate the raw position to the corresponding degree or reverse.
    """
    hw_device: ML600  # for typing's sake
    identifier = "B"


    def _change_connections(self, raw_position, reverse: bool = False) -> int:
        """
        Translate the raw position to the corresponding degree for the valve or reverse.

        Parameters:
        -----------
        raw_position : int
            The raw position value to be translated.
        reverse : bool, optional
            If True, performs the reverse translation (default is False).

        Returns:
        --------
        int
            The translated position in degrees.
        """
        if not reverse:
            translated = raw_position * 45
        else:
            translated = round(raw_position / 45)
        return translated


class ML600RightValve(ThreePortFourPositionValve):
    """
    Represents the right valve of the ML600 pump with specific translation for raw positions.

    This valve has 4 possible positions each separated by 90 degrees.

    Attributes:
    -----------
    hw_device : ML600
        The hardware device instance associated with this valve.
    identifier : str
        The identifier for this valve, set to "C".

    Methods:
    --------
    _change_connections(raw_position: int, reverse: bool = False) -> int:
        Translate the raw position to the corresponding degree or reverse.
    """
    hw_device: ML600  # for typing's sake
    identifier = "C"

    def _change_connections(self, raw_position, reverse: bool = False) -> int:
        """
        Translate the raw position to the corresponding degree for the valve or reverse.

        Parameters:
        -----------
        raw_position : int
            The raw position value to be translated.
        reverse : bool, optional
            If True, performs the reverse translation (default is False).

        Returns:
        --------
        int
            The translated position in degrees.
        """
        if not reverse:
            translated = (raw_position + 2) * 90
            if translated >= 360:
                translated -= 360
        else:
            # round, the return is often off by 1°/the valve does not switch exactly
            # the slightly complicated logic here is because the degrees are differently defined in the abstract valve
            # and the physical ML600 valve, the offset in multiples of 90 degrees is corrected here
            translated = round(raw_position / 90)
            if translated < 2:
                translated += 2
            else:
                translated -= 2

        return translated



"""Huber chiller control driver."""
import asyncio

import aioserial
import pint
from loguru import logger

from flowchem import ureg
from flowchem.components.device_info import DeviceInfo
from flowchem.components.technical.temperature import TempRange
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.huber.huber_temperature_control import HuberTemperatureControl
from flowchem.devices.huber.pb_command import PBCommand
from flowchem.utils.exceptions import InvalidConfigurationError
from flowchem.utils.people import dario, jakob, wei_hsin


class HuberChiller(FlowchemDevice):
    """Control class for Huber chillers."""

    DEFAULT_CONFIG = {
        "timeout": 0.1,
        "baudrate": 9600,
        "parity": aioserial.PARITY_NONE,
        "stopbits": aioserial.STOPBITS_ONE,
        "bytesize": aioserial.EIGHTBITS,
    }

    def __init__(
        self,
        aio: aioserial.AioSerial,
        name="",
        min_temp: float = -150,
        max_temp: float = 250,
    ) -> None:
        super().__init__(name)
        self._serial = aio
        self._min_t: float = min_temp
        self._max_t: float = max_temp

        self.device_info = DeviceInfo(
            authors=[dario, jakob, wei_hsin],
            manufacturer="Huber",
            model="generic chiller",
        )

    @classmethod
    def from_config(cls, port, name=None, **serial_kwargs):
        """Create instance from config dict. Used by server to initialize obj from config.

        Only required parameter is 'port'. Optional 'loop' + others (see AioSerial())
        """
        # Merge default settings, including serial, with provided ones.
        configuration = HuberChiller.DEFAULT_CONFIG | serial_kwargs

        try:
            serial_object = aioserial.AioSerial(port, **configuration)
        except (OSError, aioserial.SerialException) as serial_exception:
            raise InvalidConfigurationError(
                f"Cannot connect to the HuberChiller on the port <{port}>"
            ) from serial_exception

        return cls(serial_object, name)

    async def initialize(self):
        """Ensure the connection w/ device is working."""
        self.device_info.serial_number = str(await self.serial_number())
        if self.device_info.serial_number == "0":
            raise InvalidConfigurationError("No reply received from Huber Chiller!")
        logger.debug(
            f"Connected with Huber Chiller S/N {self.device_info.serial_number}",
        )

        # Validate temperature limits
        device_limits = await self.temperature_limits()
        if self._min_t < device_limits[0]:
            logger.warning(
                f"The device minimum temperature is higher than the specified minimum temperature!"
                f"The lowest possible temperature will be {device_limits[0]} °C",
            )
            self._min_t = device_limits[0]

        if self._max_t > device_limits[1]:
            logger.warning(
                f"The device maximum temperature is lower than the specified maximum temperature!"
                f"The maximum possible temperature will be {device_limits[1]} °C",
            )
            self._max_t = device_limits[1]

        temperature_range = TempRange(
            min=ureg.Quantity(f"{self._min_t} °C"),
            max=ureg.Quantity(f"{self._max_t} °C"),
        )

        # Set TemperatureControl component.
        self.components.append(
            HuberTemperatureControl("temperature-control", self, temperature_range)
        )

    async def _send_command_and_read_reply(self, command: str) -> str:
        """Send a command to the chiller and read the reply.

        Args:
        ----
            command (str): string to be transmitted

        Returns:
        -------
            str: reply received
        """
        # Send command. Using PBCommand ensure command validation, see PBCommand.to_chiller()
        pb_command = PBCommand(command.upper())
        await self._serial.write_async(pb_command.to_chiller())
        logger.debug(f"Command {command[0:8]} sent!")

        # Receive reply and return it after decoding
        try:
            reply = await asyncio.wait_for(self._serial.readline_async(), 3)
        except asyncio.TimeoutError:
            logger.error("No reply received! Unsupported command?")
            return ""

        logger.debug(f"Reply received: {reply}")
        return reply.decode("ascii")

    async def get_temperature(self) -> float:
        """Get temperature. Process preferred, otherwise internal."""
        if process_t := await self.process_temperature():
            return process_t
        return await self.internal_temperature()  # type: ignore

    async def get_temperature_setpoint(self) -> float | None:
        """Return the set point used by temperature controller. Internal if not probe, otherwise process temp."""
        reply = await self._send_command_and_read_reply("{M00****")
        return PBCommand(reply).parse_temperature()

    async def set_temperature(self, temp: pint.Quantity):
        """Set the set point used by temperature controller. Internal if not probe, otherwise process temp."""
        await self._send_command_and_read_reply("{M00" + self._temp_to_string(temp))

    async def target_reached(self) -> bool:
        """Trivially implemented as delta(currentT-setT) < 1°C."""
        current_t = await self.get_temperature()
        set_t = await self.get_temperature_setpoint()
        if set_t:
            return abs(current_t - set_t) < 1
        return False

    async def internal_temperature(self) -> float | None:
        """Return internal temp (bath temperature)."""
        reply = await self._send_command_and_read_reply("{M01****")
        return PBCommand(reply).parse_temperature()

    async def process_temperature(self) -> float | None:
        """Return the current process temperature. If not T probe, the temperature is None."""
        reply = await self._send_command_and_read_reply("{M07****")

        return PBCommand(reply).parse_temperature()

    async def temperature_limits(self) -> tuple[float, float]:
        """Return minimum/maximum accepted value for the temperature setpoint (in Celsius)."""
        min_reply = await self._send_command_and_read_reply("{M30****")
        min_t = PBCommand(min_reply).parse_temperature()
        max_reply = await self._send_command_and_read_reply("{M31****")
        max_t = PBCommand(max_reply).parse_temperature()
        return min_t, max_t

    async def serial_number(self) -> int:
        """Get serial number."""
        serial1 = await self._send_command_and_read_reply("{M1B****")
        serial2 = await self._send_command_and_read_reply("{M1C****")
        pb1, pb2 = PBCommand(serial1), PBCommand(serial2)
        if pb1.data and pb2.data:
            return int(pb1.data + pb2.data, 16)
        else:
            return 0

    @staticmethod
    def _temp_to_string(temp: pint.Quantity) -> str:
        """From temperature to string for command. f^-1 of PCommand.parse_temperature."""
        assert (
            ureg.Quantity("-151 °C") <= temp <= ureg.Quantity("327 °C")
        ), "Protocol temperature limits"
        # Hexadecimal two's complement
        return f"{int(temp.m_as('°C') * 100) & 65535:04X}"

    @staticmethod
    def _int_to_string(number: int) -> str:
        """From int to string for command. f^-1 of PCommand.parse_integer."""
        return f"{number:04X}"

    # async def return_temperature(self) -> float | None:
    #     """Return the temp of the thermal fluid flowing back to the device."""
    #     reply = await self._send_command_and_read_reply("{M02****")
    #     return PBCommand(reply).parse_temperature()
    #
    # async def pump_pressure(self) -> str:
    #     """Return the pump pressure in mbar (note that you probably want barg, i.e. to remove 1 bar)."""
    #     reply = await self._send_command_and_read_reply("{M03****")
    #     pressure = PBCommand(reply).parse_integer()
    #     return str(ureg(f"{pressure} mbar"))
    #
    # async def current_power(self) -> str:
    #     """Return the current power in Watts (negative for cooling, positive for heating)."""
    #     reply = await self._send_command_and_read_reply("{M04****")
    #     power = PBCommand(reply).parse_integer()
    #     return str(ureg(f"{power} watt"))
    #
    # async def status(self) -> dict[str, bool]:
    #     """Return the info contained in `vstatus1` as dict."""
    #     reply = await self._send_command_and_read_reply("{M0A****")
    #     return PBCommand(reply).parse_status1()
    #
    # async def status2(self) -> dict[str, bool]:
    #     """Return the info contained in `vstatus2` as dict."""
    #     reply = await self._send_command_and_read_reply("{M3C****")
    #     return PBCommand(reply).parse_status2()
    #
    # async def is_temperature_control_active(self) -> bool:
    #     """Return whether temperature control is active or not."""
    #     reply = await self._send_command_and_read_reply("{M14****")
    #     return PBCommand(reply).parse_boolean()
    #
    # async def is_circulation_active(self) -> bool:
    #     """Return whether temperature control is active or not."""
    #     reply = await self._send_command_and_read_reply("{M16****")
    #     return PBCommand(reply).parse_boolean()
    #
    # async def start_circulation(self):
    #     """Start circulation pump."""
    #     await self._send_command_and_read_reply("{M160001")
    #
    # async def stop_circulation(self):
    #     """Stop circulation pump."""
    #     await self._send_command_and_read_reply("{M160000")
    #
    # async def pump_speed(self) -> str:
    #     """Return current circulation pump speed (in rpm)."""
    #     reply = await self._send_command_and_read_reply("{M26****")
    #     return PBCommand(reply).parse_rpm()
    #
    # async def pump_speed_setpoint(self) -> str:
    #     """Return the set point of the circulation pump speed (in rpm)."""
    #     reply = await self._send_command_and_read_reply("{M48****")
    #     return PBCommand(reply).parse_rpm()
    #
    # async def set_pump_speed(self, rpm: str):
    #     """Set the pump speed, in rpm. See device display for range."""
    #     parsed_rpm = ureg(rpm)
    #     await self._send_command_and_read_reply(
    #         "{M48" + self._int_to_string(parsed_rpm.m_as("rpm"))
    #     )
    #
    # async def cooling_water_temp(self) -> float | None:
    #     """Return the cooling water inlet temperature (in Celsius)."""
    #     reply = await self._send_command_and_read_reply("{M2C****")
    #     return PBCommand(reply).parse_temperature()
    #
    # async def cooling_water_pressure(self) -> float | None:
    #     """Return the cooling water inlet pressure (in mbar)."""
    #     reply = await self._send_command_and_read_reply("{M2D****")
    #     if pressure := PBCommand(reply).parse_integer() == 64536:
    #         return None
    #     return pressure
    #
    # async def cooling_water_temp_outflow(self) -> float | None:
    #     """Return the cooling water outlet temperature (in Celsius)."""
    #     reply = await self._send_command_and_read_reply("{M4C****")
    #     return PBCommand(reply).parse_temperature()
    #
    # async def alarm_max_internal_temp(self) -> float | None:
    #     """Return the max internal temp before the alarm is triggered and a fault generated."""
    #     reply = await self._send_command_and_read_reply("{M51****")
    #     return PBCommand(reply).parse_temperature()
    #
    # async def set_alarm_max_internal_temp(self, set_temp: str):
    #     """Set the max internal temp before the alarm is triggered and a fault generated."""
    #     temp = ureg(set_temp)
    #     await self._send_command_and_read_reply("{M51" + self._temp_to_string(temp))
    #
    # async def alarm_min_internal_temp(self) -> float | None:
    #     """Return the min internal temp before the alarm is triggered and a fault generated."""
    #     reply = await self._send_command_and_read_reply("{M52****")
    #     return PBCommand(reply).parse_temperature()
    #
    # async def set_alarm_min_internal_temp(self, set_temp: str):
    #     """Set the min internal temp before the alarm is triggered and a fault generated."""
    #     temp = ureg(set_temp)
    #     await self._send_command_and_read_reply("{M52" + self._temp_to_string(temp))
    #
    # async def alarm_max_process_temp(self) -> float | None:
    #     """Return the max process temp before the alarm is triggered and a fault generated."""
    #     reply = await self._send_command_and_read_reply("{M53****")
    #     return PBCommand(reply).parse_temperature()
    #
    # async def set_alarm_max_process_temp(self, set_temp: str):
    #     """Set the max process temp before the alarm is triggered and a fault generated."""
    #     temp = ureg(set_temp)
    #     await self._send_command_and_read_reply("{M53" + self._temp_to_string(temp))
    #
    # async def alarm_min_process_temp(self) -> float | None:
    #     """Return the min process temp before the alarm is triggered and a fault generated."""
    #     reply = await self._send_command_and_read_reply("{M54****")
    #     return PBCommand(reply).parse_temperature()
    #
    # async def set_alarm_min_process_temp(self, set_temp: str):
    #     """Set the min process temp before the alarm is triggered and a fault generated."""
    #     temp = ureg(set_temp)
    #     await self._send_command_and_read_reply("{M54" + self._temp_to_string(temp))
    #
    # async def set_ramp_duration(self, ramp_time: str):
    #     """Set the duration (in seconds) of a ramp to the temperature set by a later call to ramp_to_temperature."""
    #     parsed_time = ureg(ramp_time)
    #     await self._send_command_and_read_reply(
    #         "{M59" + self._int_to_string(parsed_time.m_as("s"))
    #     )
    #
    # async def ramp_to_temperature(self, temperature: str):
    #     """Set the duration (in seconds) of a ramp to the temperature set by a later call to start_ramp()."""
    #     temp = ureg(temperature)
    #     await self._send_command_and_read_reply("{M5A" + self._temp_to_string(temp))
    #
    # async def is_venting(self) -> bool:
    #     """Whether the chiller is venting or not."""
    #     reply = await self._send_command_and_read_reply("{M6F****")
    #     return PBCommand(reply).parse_boolean()
    #
    # async def start_venting(self):
    #     """Start venting. ONLY USE DURING SETUP! READ THE MANUAL!"""
    #     await self._send_command_and_read_reply("{M6F0001")
    #
    # async def stop_venting(self):
    #     """Stop venting."""
    #     await self._send_command_and_read_reply("{M6F0000")
    #
    # async def is_draining(self) -> bool:
    #     """Whether the chiller is venting or not."""
    #     reply = await self._send_command_and_read_reply("{M70****")
    #     return PBCommand(reply).parse_boolean()
    #
    # async def start_draining(self):
    #     """Start venting. ONLY USE DURING SHUT DOWN! READ THE MANUAL!"""
    #     await self._send_command_and_read_reply("{M700001")
    #
    # async def stop_draining(self):
    #     """Stop venting."""
    #     await self._send_command_and_read_reply("{M700000")


if __name__ == "__main__":
    device = HuberChiller(aioserial.AioSerial(port="COM8"))

    async def main(chiller):
        await chiller.initialize()
        print(f"S/N is {await chiller.serial_number()}")

    asyncio.run(main(device))



from typing import TYPE_CHECKING

from pydantic import AnyHttpUrl

if TYPE_CHECKING:
    from flowchem.client.device_client import FlowchemDeviceClient
from flowchem.components.component_info import ComponentInfo


class FlowchemComponentClient:
    def __init__(self, url: AnyHttpUrl, parent: "FlowchemDeviceClient") -> None:
        self.base_url = str(url)
        self._parent = parent
        self._session = parent._session
        self.component_info = ComponentInfo.model_validate_json(self.get("").text)

    def get(self, url, **kwargs):
        """Send a GET request. Returns :class:`Response` object."""
        return self._session.get(self.base_url + "/" + url, **kwargs)

    def post(self, url, data=None, json=None, **kwargs):
        """Send a POST request. Returns :class:`Response` object."""
        return self._session.post(
            self.base_url + "/" + url, data=data, json=json, **kwargs
        )

    def put(self, url, data=None, **kwargs):
        """Send a PUT request. Returns :class:`Response` object."""

        # Inspect the keyword arguments type to avoid problems with arguments of list type which must be handed to fastapi as str
        for key, arg in kwargs.get("params", {}).items():
            if type(arg) is list:
                kwargs["params"][key] = str(arg)
        
        return self._session.put(self.base_url + "/" + url, data=data, **kwargs)



"""Base pump component."""
from flowchem.components.flowchem_component import FlowchemComponent
from flowchem.devices.flowchem_device import FlowchemDevice


class Pump(FlowchemComponent):
    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        """A generic pump."""
        super().__init__(name, hw_device)
        self.add_api_route("/infuse", self.infuse, methods=["PUT"])
        self.add_api_route("/stop", self.stop, methods=["PUT"])
        self.add_api_route("/is-pumping", self.is_pumping, methods=["GET"])
        if self.is_withdrawing_capable():
            self.add_api_route("/withdraw", self.withdraw, methods=["PUT"])
        self.component_info.type = "Pump"

    async def infuse(self, rate: str = "", volume: str = "") -> bool:  # type: ignore
        """Start infusion."""
        ...

    async def stop(self):  # type: ignore
        """Stop pumping."""
        ...

    async def is_pumping(self) -> bool:  # type: ignore
        """Is pump running?"""
        ...

    @staticmethod
    def is_withdrawing_capable() -> bool:  # type: ignore
        """Can the pump reverse its normal flow direction?"""
        ...

    async def withdraw(self, rate: str = "", volume: str = "") -> bool:  # type: ignore
        """Pump in the opposite direction of infuse."""
        ...



"""Syringe pump component, two flavours, infuse only, infuse-withdraw."""

from flowchem.components.pumps.pump import Pump
from flowchem.devices.flowchem_device import FlowchemDevice


class HPLCPump(Pump):
    def __init__(self, name: str, hw_device: FlowchemDevice) -> None:
        super().__init__(name, hw_device)

        # Ontology: HPLC isocratic pump
        self.component_info.owl_subclass_of.append(
            "http://purl.obolibrary.org/obo/OBI_0000556",
        )
        self.component_info.type = "HPLC Pump"

    @staticmethod
    def is_withdrawing_capable() -> bool:
        """HPLC pump cannot reverse flow direction (fundamental limit due to displacement + check valves)."""
        return False



"""Control Hamilton ML600 syringe pump via the protocol1/RNO+."""
from __future__ import annotations

import asyncio
import string
import warnings
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING

import aioserial
from loguru import logger

from flowchem import ureg
from flowchem.components.device_info import DeviceInfo
from flowchem.devices.flowchem_device import FlowchemDevice
from flowchem.devices.hamilton.ml600_pump import ML600Pump
from flowchem.devices.hamilton.ml600_valve import ML600LeftValve, ML600RightValve
from flowchem.utils.exceptions import InvalidConfigurationError, DeviceError
from flowchem.utils.people import dario, jakob, wei_hsin

if TYPE_CHECKING:
    import pint


# i.e. PUMP_ADDRESS = {1: 'a', 2: 'b', 3: 'c', 4: 'd', ..., 16: 'p'}
# Note ':' is used for broadcast within the daisy chain.
PUMP_ADDRESS = dict(enumerate(string.ascii_lowercase[:16], start=1))


@dataclass
class Protocol1Command:
    """Class representing a pump command and its expected reply."""

    command: str
    target_pump_num: int = 1
    target_component: str = ""
    command_value: str = ""
    optional_parameter: str = ""
    parameter_value: str = ""
    execution_command: str = "R"  # Execute

    def compile(self) -> str:
        """Create actual command byte by prepending pump address to command and appending executing command."""
        compiled_command = (
            f"{PUMP_ADDRESS[self.target_pump_num]}"
            f"{self.target_component}"
            f"{self.command}{self.command_value}"
        )

        if self.parameter_value:
            compiled_command += f"{self.optional_parameter}{self.parameter_value}"

        return compiled_command + self.execution_command


class HamiltonPumpIO:
    """Setup with serial parameters, low level IO."""

    ACKNOWLEDGE = chr(6)
    NEGATIVE_ACKNOWLEDGE = chr(21)
    DEFAULT_CONFIG = {
        "timeout": 0.1,
        "baudrate": 9600,
        "parity": aioserial.PARITY_ODD,
        "stopbits": aioserial.STOPBITS_ONE,
        "bytesize": aioserial.SEVENBITS,
    }

    def __init__(self, aio_port: aioserial.Serial) -> None:
        """Initialize serial port, not pumps."""
        self._serial = aio_port
        self.num_pump_connected: int | None = (
            None  # Set by `HamiltonPumpIO.initialize()`
        )

    @classmethod
    def from_config(cls, config):
        """Create HamiltonPumpIO from config."""
        configuration = HamiltonPumpIO.DEFAULT_CONFIG | config

        try:
            serial_object = aioserial.AioSerial(**configuration)
        except aioserial.SerialException as serial_exception:
            raise InvalidConfigurationError(
                f"Cannot connect to the pump on the port <{configuration.get('port')}>"
            ) from serial_exception

        return cls(serial_object)

    async def initialize(self, hw_initialization: bool = True):
        """Ensure connection with pump and initialize it (if hw_initialization is True)."""
        self.num_pump_connected = await self._assign_pump_address()
        if hw_initialization:
            await self.all_hw_init()  # initialization take more than 8.5 sec for one instrument
            await asyncio.sleep(8)  # this might be necessary due to checking request_done sometime fail with "" return

    async def _assign_pump_address(self) -> int:
        """Auto assign pump addresses.

        To be run on init, auto assign addresses to pumps based on their position in the daisy chain.
        A custom command syntax with no addresses is used here so read and write has been rewritten
        """
        try:
            # the command for an unknown reason often replies wrongly on first attempt. therefore, this is done twice
            await self._write_async(b"1a\r")
            await self._read_reply_async()
            await self._write_async(b"1a\r")
        except aioserial.SerialException as e:
            raise InvalidConfigurationError from e

        reply = await self._read_reply_async()
        if not reply or reply[:1] != "1":
            raise InvalidConfigurationError(f"No pump found on {self._serial.port}")
        # reply[1:2] should be the address of the last pump. However, this does not work reliably.
        # So here we enumerate the pumps explicitly instead
        last_pump = 0
        for pump_num, address in PUMP_ADDRESS.items():
            await self._write_async(f"{address}UR\r".encode("ascii"))
            if "NV01" in await self._read_reply_async():
                last_pump = pump_num
            else:
                break
        logger.debug(f"Found {last_pump} pumps on {self._serial.port}!")
        return int(last_pump)

    async def all_hw_init(self):
        """Send to all pumps the HW initialization command (i.e. homing)."""
        await self._write_async(b":K\r")
        await self._write_async(b":V\r")
        await self._write_async(b":#SP2\r")
        await self._write_async(b":XR\r")  # Broadcast: initialize + execute
        # Note: no need to consume reply here because there is none (since we are using broadcast)

    async def _write_async(self, command: bytes):
        """Write a command to the pump."""
        await self._serial.write_async(command)
        logger.debug(f"Command {command!r} sent!")

    async def _read_reply_async(self) -> str:
        """Read the pump reply from serial communication."""
        reply_string = await self._serial.readline_async()
        logger.debug(f"Reply received: {reply_string}")
        return reply_string.decode("ascii")

    def _parse_response(self, response: str) -> str:
        """Split a received line in its components: status, reply."""
        status, reply = response[:1], response[1:]

        assert status in (self.ACKNOWLEDGE, self.NEGATIVE_ACKNOWLEDGE)
        if status == self.NEGATIVE_ACKNOWLEDGE:
            logger.warning("Negative acknowledge received")
            warnings.warn(
                "Negative acknowledge reply: check command syntax!",
                stacklevel=2,
            )

        return reply.rstrip()  # removes trailing <cr>

    def _translate_ascii_to_binary(self, reply: str):
        binary_representation = ''.join(format(byte, '08b') for byte in reply.encode('ascii'))[::-1]
        return binary_representation

        # binary_list = []
        # [binary_list.append(format(byte, '08b')[::-1]) for byte in reply.encode('ascii')]
        # all_status = binary_list[0]

    async def write_and_read_reply_async(self, command: Protocol1Command) -> str:
        """Send a command to the pump, read the replies and returns it, optionally parsed."""
        self._serial.reset_input_buffer()
        await self._write_async(f"{command.compile()}\r".encode("ascii"))
        response = await self._read_reply_async()

        if not response:
            raise InvalidConfigurationError(
                f"No response received from pump! "
                f"Maybe wrong pump address? (Set to {command.target_pump_num})"
            )

        return self._parse_response(response)


class ML600Commands(Enum):
    """ Just a collection of commands. Grouped here to ease future, unlikely, changes. """

    PAUSE = "K"
    RESUME = "$"
    CLEAR_BUFFER = "V"

    INIT_ALL = "X"
    INIT_VALVE_ONLY = "LX"
    INIT_SYRINGE_ONLY = "X1"

    # only works for pumps with two syringe drivers
    SET_VALVE_CONTINUOUS_DISPENSE = "LST19"
    SET_VALVE_DUAL_DILUTOR = "LST20"

    # if there are two drivers, both sides can be selected
    SELECT_LEFT = "B"
    SELECT_RIGHT = "C"

    # SYRINGE POSITION
    PICKUP = "P"
    DELIVER = "D"
    ABSOLUTE_MOVE = "M"

    # VALVE POSITION
    # strongly discouraged since mapping changes
    VALVE_TO_INLET = "I"
    VALVE_TO_OUTLET = "O"
    VALVE_TO_WASH = "W"
    VALVE_BY_NAME_CW = "LP0"
    VALVE_BY_NAME_CCW = "LP1"
    # strongly encouraged since mapping is clear if initial/0 position is clear and rotor/stator are known
    VALVE_BY_ANGLE_CW = "LA0"
    VALVE_BY_ANGLE_CCW = "LA1"

    # STATUS REQUEST
    # INFORMATION REQUEST -- these all returns Y/N/* where * means busy
    REQUEST_DONE = "F"
    SYRINGE_HAS_ERROR = "Z"
    VALVE_HAS_ERROR = "G"
    IS_SINGLE_SYRINGE = "H"
    # STATUS REQUEST  - these have complex responses, see relevant methods for details.
    STATUS_REQUEST = "E1"
    ERROR_REQUEST = "E2"
    TIMER_REQUEST = "E3"
    BUSY_STATUS = "T1"
    ERROR_STATUS = "T2"
    # PARAMETER REQUEST
    SYRINGE_DEFAULT_SPEED = "YQS"
      # 2-3692 seconds per stroke
    CURRENT_SYRINGE_POSITION = "YQP"  # 0-52800 steps
    SYRINGE_DEFAULT_BACKOFF = "YQB"  # 0-1000 steps
    CURRENT_VALVE_POSITION = "LQP"
      # 1-8 (see docs, Table 3.2.2
    GET_RETURN_STEPS = "YQN"  # 0-1000 steps
    # PARAMETER CHANGE
    SET_RETURN_STEPS = "YSN"  # 0-1000
    # VALVE REQUEST
    VALVE_ANGLE = "LQA"  # 0-359 degrees
    VALVE_CONFIGURATION = "YQS"
      # 11-20 (see docs, Table 3.2.2
    #Set valve speed
    SET_VALVE_SPEED = "LSF"  # 15-720 degrees per sec
    #Set valve speed
    GET_VALVE_SPEED = "LQF"
    # TIMER REQUEST
    TIMER_DELAY = "<T"  # 0–99999999 ms
    # FIRMWARE REQUEST
    FIRMWARE_VERSION = "U"
      # xxii.jj.k (ii major, jj minor, k revision)
    OPTIONAL_PARAMETER = "S"

class ML600(FlowchemDevice):
    """ML600 implementation according to manufacturer docs. Tested on a 61501-01 (i.e. single syringe system).

    From manufacturer docs:
    To determine the volume dispensed per step the total syringe volume is divided by
    48,000 steps. All Hamilton instrument syringes are designed with a 60 mm stroke
    length and the Microlab 600 is designed to move 60 mm in 48,000 steps. For
    example to dispense 9 mL from a 10 mL syringe you would determine the number of
    steps by multiplying 48000 steps (9 mL/10 mL) to get 43,200 steps.
    """

    DEFAULT_CONFIG = {
        "default_infuse_rate": "1 ml/min",
        "default_withdraw_rate": "1 ml/min",
    }

    # This class variable is used for daisy chains (i.e. multiple pumps on the same serial connection). Details below.
    _io_instances: set[HamiltonPumpIO] = set()
    # The mutable object (a set) as class variable creates a shared state across all the instances.
    # When several pumps are daisy-chained on the same serial port, they need to all access the same Serial object,
    # because access to the serial port is exclusive by definition (also locking there ensure thread safe operations).

    # Only Hamilton syringes are compatible w/ the ML600, and they come on a limited set of sizes. (Values in ml)
    VALID_SYRINGE_VOLUME = {
        0.01,
        0.025,
        0.05,
        0.1,
        0.25,
        0.5,
        1.0,
        2.5,
        5.0,
        10.0,
        25.0,
        50.0,
    }

    def __init__(
        self,
        pump_io: HamiltonPumpIO,
        syringe_volume: str,
        name: str,
        address: int = 1,
        **config,
    ) -> None:
        """Default constructor, needs an HamiltonPumpIO object. See from_config() class method for config-based init.

        Args:
        ----
            pump_io: An HamiltonPumpIO w/ serial connection to the daisy chain w/ target pump.
            syringe_volume: Volume of the syringe used, either a Quantity or number in ml.
            address: number of pump in array, 1 for first one, auto-assigned on init based on position.
            name: 'cause naming stuff is important.
        """
        super().__init__(name)
        self.device_info = DeviceInfo(
            authors=[dario, jakob, wei_hsin],
            manufacturer="Hamilton",
            model="ML600",
        )
        # HamiltonPumpIO
        self.pump_io = pump_io
        ML600._io_instances.add(self.pump_io)  # See above for details.

        # Pump address is the pump sequence number if in chain. Count starts at 1, default.
        self.address = int(address)

        # Syringe pumps only perform linear movement, and the volume displaced is function of the syringe loaded.
        try:
            # todo: set syringe_volume to a dict/str??
            self.syringe_volume = ureg.Quantity(syringe_volume)
        except AttributeError as attribute_error:
            logger.error(f"Invalid syringe volume {syringe_volume}!")
            raise InvalidConfigurationError(
                "Invalid syringe volume provided."
                "The syringe volume is a string with units! e.g. '5 ml'"
            ) from attribute_error

        if self.syringe_volume.m_as("ml") not in ML600.VALID_SYRINGE_VOLUME:
            raise InvalidConfigurationError(
                f"The specified syringe volume ({syringe_volume}) is invalid!\n"
                f"The volume (in ml) has to be one of {ML600.VALID_SYRINGE_VOLUME}"
            )

        self._steps_per_ml = ureg.Quantity(f"{48000 / self.syringe_volume} step")
        # todo: check
        # self._offset_steps = 100  # Steps added to each absolute move command, to decrease wear and tear at volume = 0
        # self._max_vol = (48000 - self._offset_steps) * ureg.step / self._steps_per_ml
        self.return_steps = 24  # Steps added to each absolute move command (default)

        # This enables to configure on per-pump basis uncommon parameters
        self.config = ML600.DEFAULT_CONFIG | config
        self.dual_syringe = False

    @classmethod
    def from_config(cls, **config):
        """Create instances via config file."""
        # Many pump can be present on the same serial port with different addresses.
        # This shared list of HamiltonPumpIO objects allow shared state in a borg-inspired way, avoiding singletons
        # This is only relevant to programmatic instantiation, i.e. when from_config() is called per each pump from a
        # config file, as it is the case in the HTTP server.
        pumpio = None
        for obj in ML600._io_instances:
            # noinspection PyProtectedMember
            if obj._serial.port == config.get("port"):
                pumpio = obj
                break

        # If not existing serial object are available for the port provided, create a new one
        if pumpio is None:
            # Remove ML600-specific keys to only have HamiltonPumpIO's kwargs
            config_for_pumpio = {
                k: v
                for k, v in config.items()
                if k not in ("syringe_volume", "address", "name")
            }
            pumpio = HamiltonPumpIO.from_config(config_for_pumpio)

        return cls(
            pumpio,
            syringe_volume=config.get("syringe_volume", ""),
            address=config.get("address", 1),
            name=config.get("name", ""),
        )

    async def get_return_steps(self) -> int:
        """ Gives the defined return steps for syringe movement """
        reply=await self.send_command_and_read_reply(Protocol1Command(command=ML600Commands.GET_RETURN_STEPS.value))
        return int(reply)

    async def set_return_steps(self, return_steps: int):
        """
        Return steps are used to compensate for the mechanical drive system backlash,
        independent of syringe size and default is 24 steps.
        The return step should be in # 0-1000 steps.
        """
        # waiting is necessary since this happens on (class) initialisation
        target_steps = str(int(return_steps))
        await self.wait_until_system_idle()
        await self.send_command_and_read_reply(Protocol1Command(
            command=ML600Commands.SET_RETURN_STEPS.value, command_value=target_steps))

    async def initialize(self, hw_init=False, init_speed: str = "200 sec / stroke"):
        """Initialize pump and its components."""
        # this command MUST be executed in the beginning
        await self.pump_io.initialize()
        await self.wait_until_system_idle()
        # Test connectivity by querying the pump's firmware version
        self.device_info.version = await self.version()
        logger.info(
            f"Connected to Hamilton ML600 {self.name} - FW version: {self.device_info.version}!",
        )
        self.dual_syringe = not await self.is_single_syringe()
        await self.general_status_info()

        # Add device components
        if self.dual_syringe:
            self.components.extend([ML600Pump("left_pump", self, "B"), ML600Pump("right_pump", self, "C"),
                                    ML600LeftValve("left_valve", self), ML600RightValve("right_valve", self)])
        else:
            self.components.extend([ML600Pump("pump", self), ML600LeftValve("valve", self)])

    async def send_command_and_read_reply(self, command: Protocol1Command) -> str:
        """Send a command to the pump. Here we just add the right pump number."""
        command.target_pump_num = self.address
        return await self.pump_io.write_and_read_reply_async(command)

    def _validate_speed(self, speed: pint.Quantity | None) -> str:
        """Validate the speed.

        Given a speed (seconds/stroke) returns a valid value for it, and a warning if out of bounds.
        """
        # Validated speeds are used as command argument, with empty string being the default for None
        if speed is None:
            return ""

        # Alert if out of bounds but don't raise exceptions, according to general philosophy.
        # Target flow rate too high
        if speed < ureg.Quantity("2 sec/stroke"):
            speed = ureg.Quantity("2 sec/stroke")
            warnings.warn(
                f"Desired speed ({speed}) is unachievable!"
                f"Set to {self._seconds_per_stroke_to_flowrate(speed)}"
                f"Wrong units? A bigger syringe is needed?",
                stacklevel=2,
            )

        # Target flow rate too low
        if speed > ureg.Quantity("3692 sec/stroke"):
            speed = ureg.Quantity("3692 sec/stroke")
            warnings.warn(
                f"Desired speed ({speed}) is unachievable!"
                f"Set to {self._seconds_per_stroke_to_flowrate(speed)}"
                f"Wrong units? A smaller syringe is needed?",
                stacklevel=2,
            )

        return str(round(speed.m_as("sec / stroke")))

    async def initialize_pump(self, speed: pint.Quantity | None = None):
        """Initialize both syringe and valve. speed: 2-3692 in seconds/stroke"""
        init_pump = Protocol1Command(
            command=ML600Commands.INIT_ALL.value,
            optional_parameter=ML600Commands.OPTIONAL_PARAMETER.value,
            parameter_value=self._validate_speed(speed),
        )
        return await self.send_command_and_read_reply(init_pump)

    async def initialize_valve(self):
        """Initialize valve only."""
        return await self.send_command_and_read_reply(Protocol1Command(command="LX"))

    async def initialize_syringe(self, speed: pint.Quantity | None = None):
        """Initialize syringe only. speed: 2-3692 in seconds/stroke"""
        init_syringe = Protocol1Command(
            command="X1",
            optional_parameter="S",
            parameter_value=self._validate_speed(speed),
        )
        return await self.send_command_and_read_reply(init_syringe)

    def _flowrate_to_seconds_per_stroke(self, flowrate: pint.Quantity):
        """Convert flow rates to steps per seconds.

        To determine the volume dispensed per step the total syringe volume is divided by
        48,000 steps. All Hamilton instrument syringes are designed with a 60 mm stroke
        length and the Microlab 600 is designed to move 60 mm in 48,000 steps. For
        example to dispense 9 mL from a 10 mL syringe you would determine the number of
        steps by multiplying 48000 steps (9 mL/10 mL) to get 43,200 steps.
        """
        flowrate_in_steps_sec = flowrate * self._steps_per_ml
        return (1 / flowrate_in_steps_sec).to("second/stroke")

    def _seconds_per_stroke_to_flowrate(self, second_per_stroke) -> float:
        """Convert seconds per stroke to flow rate."""
        flowrate = 1 / (second_per_stroke * self._steps_per_ml)
        return flowrate.to("ml/min")

    def _volume_to_step_position(self, volume: pint.Quantity) -> int:
        """Convert a volume to a step position."""
        # todo: different syringes
        # noinspection PyArgumentEqualDefault
        steps = volume * self._steps_per_ml
        return round(steps.m_as("steps"))

    async def get_current_volume(self, pump: str) -> pint.Quantity:
        """Return current syringe position in ml."""
        syringe_pos = await self.send_command_and_read_reply(
            Protocol1Command(command=ML600Commands.CURRENT_SYRINGE_POSITION.value,
                             target_component=pump),
        )

        current_steps = int(syringe_pos) * ureg.step
        return current_steps / self._steps_per_ml

    async def set_to_volume(self, target_volume: pint.Quantity, rate: pint.Quantity, pump: str):
        """Absolute move to target volume provided by set step position and speed."""
        speed = self._flowrate_to_seconds_per_stroke(rate)  # in seconds/stroke
        set_speed = self._validate_speed(speed)
        position = self._volume_to_step_position(target_volume)
        logger.debug(f"Pump {self.name} set to volume {target_volume} at speed {set_speed}")

        abs_move_cmd = Protocol1Command(
            command=ML600Commands.ABSOLUTE_MOVE.value,
            optional_parameter=ML600Commands.OPTIONAL_PARAMETER.value,
            command_value=str(position),
            parameter_value=set_speed,
            target_component=pump
        )
        return await self.send_command_and_read_reply(abs_move_cmd)

    async def pause(self, pump: str):
        """Pause any running command."""
        return await self.send_command_and_read_reply(
            Protocol1Command(command="", target_component=pump, execution_command="K"),)

    async def resume(self, pump: str):
        """Resume any paused command."""
        return await self.send_command_and_read_reply(
            Protocol1Command(command="", target_component=pump, execution_command="$"),)

    async def stop(self, pump: str) -> bool:
        """Stop and abort any running command."""
        await self.pause(pump)
        await self.send_command_and_read_reply(
            Protocol1Command(command="", target_component=pump, execution_command="V"),)
        return True  # Todo: need?

    async def get_pump_status(self, pump: str = "") -> bool:
        """Ture means pump is busy. False means pump is idle."""
        checking_mapping = {"B": 1, "C": 3}
        pump = "B" if not pump else pump
        status = await self.get_component_status(checking_mapping[pump])
        logger.info(f"pump {pump} is busy: {status}")
        return status

    async def get_valve_status(self, valve: str = "") -> bool:
        """Ture means valve is busy. False means valve is idle."""
        checking_mapping = {"B": 0, "C": 2}
        valve = "B" if not valve else valve
        status = await self.get_component_status(checking_mapping[valve])
        logger.info(f"valve {valve} is busy: {status}")
        return status

    async def get_all_component_status(self) -> dict[str, bool]:

        reply = await self.send_command_and_read_reply(
            Protocol1Command(command="T1", execution_command=""))
        all_status = ''.join(format(byte, '08b') for byte in reply.encode('ascii'))[::-1]

        value_map = {0: "left_valve busy", 1: "left_pump busy",
                     2: "right_valve busy", 3: "right_pump busy",
                     4: "step_active busy", 5: "handprobe_active busy"}
        status = {}
        for key in value_map:
            logger.debug(f"{value_map[key]} : {all_status[key] == '0'}")
            status[value_map[key]] = all_status[key] == "0"
        return status

    async def get_component_status(self, component: int = -1) -> bool:
        """
        Represent the status of specific component. True means busy; False meaens idle.
        Return status of all parts of instrument in dictionary.
        """
        reply = await self.send_command_and_read_reply(
                Protocol1Command(command="T1", execution_command=""))
        all_status = ''.join(format(byte, '08b') for byte in reply.encode('ascii'))[::-1]
        # 1 is true and 0 is false according to the manual; but the real signal is opposite.
        return all_status[component] == "0"

    async def general_status_info(self, component: int = -1) -> bool | dict[str, bool]:
        # todo: this command will reset the error of syntax and instrument, use others error monitoring method
        reply = await self.send_command_and_read_reply(
            Protocol1Command(command="E1", execution_command=""))
        binary_representation = ''.join(format(byte, '08b') for byte in reply.encode('ascii'))[::-1]

        if -1 < component < 5:
            # for specific component
            return binary_representation[component] == "1"

        value_map = {0: "instrument idle, command buffer isn't empty",
                     1: "syringe(s) busy",
                     2: "valve(s) busy",
                     3: "syntax error",
                     4: "valve or syringe error"}
        status = {}
        for key in value_map:
            status[value_map[key]] = binary_representation[key] == '1'
            logger.info(f"{value_map[key]} : {status[value_map[key]]}")


            if status[value_map[key]]:
                raise DeviceError((
                    f"{value_map[key]} shows {status[value_map[key]]}. Check! "
                ))
        return status

    async def wait_until_system_idle(self):
        """Return when no more commands are present in the pump buffer."""
        logger.debug(f"ML600 {self.name} wait until idle...")
        while not await self.is_system_idle():
            await asyncio.sleep(0.1)
        logger.debug(f"...ML600 {self.name} idle now!")

    async def is_system_idle(self) -> bool:
        """Check if the pump is idle (actually check if the last command has ended)."""
        return (
            await self.send_command_and_read_reply(
                Protocol1Command(command="F", execution_command="")) == "Y"
        )

    async def is_single_syringe(self) -> bool:
        """Determine if single or dual syringe"""
        is_single = await self.send_command_and_read_reply(
            Protocol1Command(command="H", execution_command=""),
        )
        if is_single == "N":
            return False
        elif is_single == "Y":
            return True
        else:
            raise InvalidConfigurationError("Neither single nor dual syringe - somethings wrong")

    async def version(self) -> str:
        """Return the current firmware version reported by the pump."""
        return await self.send_command_and_read_reply(Protocol1Command(command=ML600Commands.FIRMWARE_VERSION.value))

    async def is_idle(self) -> bool:
        """Check if the pump is idle (actually check if the last command has ended)."""
        return (
            await self.send_command_and_read_reply(Protocol1Command(command=ML600Commands.REQUEST_DONE.value)) == "Y"
        )

    async def get_valve_position_by_name(self, valve: ML600Commands) -> str:
        """
        Represent the position of the valve: getter returns Enum, setter needs Enum.
        Strongly encouraged to use switching by angle
        """
        return await self.send_command_and_read_reply(
            Protocol1Command(command=ML600Commands.CURRENT_VALVE_POSITION.value,
                             target_component = valve.value if self.dual_syringe else ""))

    async def set_valve_position_by_name(
        self,
        valve: ML600Commands,
        target_position: str,
        wait_for_movement_end: bool = True
    ):
        """Set valve position.
        Strongly encouraged to use switching by angle
        wait_for_movement_end is defaulted to True as it is a common mistake not to wait...
        """
        await self.send_command_and_read_reply(
            Protocol1Command(command=ML600Commands.VALVE_BY_NAME_CW.value, command_value=target_position,
                             target_component=valve.value if self.dual_syringe else ""),
        )
        logger.debug(f"{self.name} valve position set to position {target_position}")
        if wait_for_movement_end:
            while await self.get_valve_status(target_position):
                await asyncio.sleep(0.1)
        return True
            # todo: it's will be good check only pump but not whole system

    async def get_raw_position(self, target_component: str) -> str:
        """
        Represent the position of the valve: getter returns Enum, setter needs Enum.
        Strongly encouraged to use switching by angle
        """
        return await self.send_command_and_read_reply(
            Protocol1Command(command=ML600Commands.VALVE_ANGLE.value,
                             target_component=target_component if self.dual_syringe else ""))

    async def set_raw_position(
        self,
        target_position: str,
        wait_for_movement_end: bool = True,
        counter_clockwise = False,
        target_component = None
    ):
        """Set valve position.
        Strongly encouraged to use switching by angle
        wait_for_movement_end is defaulted to True as it is a common mistake not to wait...
        """
        if not counter_clockwise:
            await self.send_command_and_read_reply(
                Protocol1Command(command=ML600Commands.VALVE_BY_ANGLE_CW.value, command_value=target_position,
                                 target_component=target_component if self.dual_syringe else ""),
            )
            logger.debug(f"{self.name} valve position set to position {target_position}, switching CW")
        else:
            await self.send_command_and_read_reply(
                Protocol1Command(command=ML600Commands.VALVE_BY_ANGLE_CCW.value, command_value=target_position,
                                 target_component=target_component if self.dual_syringe else ""),
            )
            logger.debug(f"{self.name} valve position set to position {target_position}, switching CCW")
        if wait_for_movement_end:
            while await self.get_valve_status(target_component):
                await asyncio.sleep(0.1)
        return True
            # todo: it's will be good check only pump but not whole system

if __name__ == "__main__":
    import asyncio

    conf = {
        "port": "COM12",
        "address": 1,
        "name": "test1",
        "syringe_volume": "5 mL",
    }
    pump1 = ML600.from_config(**conf)
    asyncio.run(pump1.initialize_pump())
